//
//  moments.cpp
//  moments
//
//  Created by Vincent Mathieu on 20/09/2019.
//  Copyright © 2019 Vincent Mathieu. All rights reserved.
//

#include "moments.hpp"


double moments_H0_SPD(int L, int M, std::complex <double> *waves)
{
  /* return ^(e)H0(LM) for waves of refectivity (e)
   * one needs to sum over the reflectivity to get the full moments.
   *      H0(LM) = ^(+)H0(LM) + ^(-)H0(LM)
   *
   * Waves ordering:
   * waves[0]         = S_0 ;
   * waves[1,2,3]     = P_{-1,0,+1} ;
   * waves[4,5,6,7,8] = D_{-2,-1,0,+1,+2}
   */
  
  double mom = 0;
  std::complex <double> tmp = 0;
  std::complex <double> S0 = waves[0];
  std::complex <double> P1m = waves[1], P0  = waves[2], P1 = waves[3];
  std::complex <double> D2m = waves[4], D1m = waves[5], D0 = waves[6],
  D1  = waves[7], D2  = waves[8];
  
  switch (10*L+M) {
      // L = 0 ----------------------------------------------------------------------------
    case 0:
      mom  = pow(abs(S0),2.) + pow(abs(P1m),2.) + pow(abs(P0),2.);
      mom += pow(abs(P1),2.) + pow(abs(D2m),2.) + pow(abs(D1m),2.);
      mom += pow(abs(D0),2.) + pow(abs(D1),2.)  + pow(abs(D2),2.);
      mom *= 2.;
      break;
      // L = 1 ----------------------------------------------------------------------------
    case 10:
      tmp  = 4./sqrt(3.)*P0*conj(S0+2./sqrt(5.)*D0);
      tmp += 4./sqrt(5.)*(P1*conj(D1) + P1m*conj(D1m));
      mom  = real(tmp);
      break;
    case 11:
      tmp  = 2./sqrt(3.)*(P1-P1m)*conj(S0-D0/sqrt(5.));
      tmp += 2.*sqrt(2./5.)*(P1*conj(D2) - P1m*conj(D2m));
      tmp += 2.*sqrt(1./5.)*(P0*conj(D1) - P0*conj(D1m));
      mom  = real(tmp);
      break;
      // L = 2 ----------------------------------------------------------------------------
    case 20:
      mom =  4./sqrt(5.)*real(S0*conj(D0));
      mom += 4./7.*pow(abs(D0),2) + 4./5.*pow(abs(P0),2);
      mom -= 2./5.*pow(abs(P1),2) + 2./5.*pow(abs(P1m),2);
      mom += 2./7.*pow(abs(D1),2) + 2./7.*pow(abs(D1m),2);
      mom -= 4./7.*pow(abs(D2),2) + 4./7.*pow(abs(D2m),2);
      break;
    case 21:
      tmp  = 2.*( D0/7. + S0/sqrt(5.) )*( conj(D1 - D1m) );
      tmp += 2./7.*sqrt(6.)*(D1*conj(D2) - D1m*conj(D2m));
      tmp += 2.*sqrt(3.)/5.*P0*conj(P1-P1m);
      mom  = real(tmp);
      break;
    case 22:
      tmp  = 2./35.*(7*sqrt(5.)*S0-10.*D0)*conj(D2+D2m);
      tmp -= 2./7.*sqrt(6.)*(D1*conj(D1m)) + 2.*sqrt(6.)/5.*P1*conj(P1m);
      mom  = real(tmp);
      break;
      // L = 3 ----------------------------------------------------------------------------
    case 30:
      tmp  = 12./7./sqrt(5.)*( sqrt(3.)*D0*conj(P0) - P1*conj(D1) - P1m*conj(D1m) );
      mom  = real(tmp);
      break;
    case 31:
      tmp  = 4./7.*sqrt(6./5.)*P0*conj(D1-D1m) + 6./7.*sqrt(2./5.)*D0*conj(P1-P1m);
      tmp -= 2./7.*sqrt(3./5.)*(P1*conj(D2) - P1m*conj(D2m));
      mom  = real(tmp);
      break;
    case 32:
      tmp  = sqrt(12.)/7.*( P0*conj(D2+D2m) - sqrt(2.)*(P1*conj(D1m) + P1m*conj(D1)) );
      mom  = real(tmp);
      break;
    case 33:
      mom  = 6./7.*real(P1*conj(D2m) - P1m*conj(D2));
      break;
      // L = 4 ----------------------------------------------------------------------------
    case 40:
      mom  = 4./7.*pow(abs(D0),2);
      mom -= 8./21.*pow(abs(D1),2) + 8./21.*pow(abs(D1m),2);
      mom += 2./21.*pow(abs(D2),2) + 2./21.*pow(abs(D2m),2);
      break;
    case 41:
      mom  = 2.*sqrt(30.)/21.*real(D0*conj( D1 - D1m ));
      mom += 2.*sqrt(5.)/21.*real(D1m*conj(D2m) - D1*conj(D2));
      break;
    case 42:
      mom  = 2.*sqrt(15.)/21.*real(D0*conj( D2+D2m ));
      mom -= 4.*sqrt(10.)/21.*real(D1*conj(D1m));
      break;
    case 43:
      mom = 2./3.*sqrt(5./7.)*real(D2m*conj(D1) - D2*conj(D1m));
      break;
    case 44:
      mom = 2./3.*sqrt(10./7.)*real(D2*conj(D2m));
      break;
    default:
      mom = 0;
      break;
  }
  
  return mom;
}



double moments_H1_SPD(int L, int M, std::complex <double> *waves)
{
  /* return H1(LM) for one set of reflectivity waves
   * one needs to SUBTRACT the two reflectivities to get the full moments
   *      H1(LM) = ^(+)H1(LM) - ^(-)H1(LM)
   *
   * Waves ordering:
   * waves[0]         = S_0 ;
   * waves[1,2,3]     = P_{-1,0,+1} ;
   * waves[4,5,6,7,8] = D_{-2,-1,0,+1,+2}
   */
  
  double mom = 0;
  std::complex <double> tmp = 0;
  std::complex <double> S0 = waves[0];
  std::complex <double> P1m = waves[1], P0  = waves[2], P1 = waves[3];
  std::complex <double> D2m = waves[4], D1m = waves[5], D0 = waves[6],
  D1  = waves[7], D2  = waves[8];
  
  switch (10*L+M) {
      // L = 0 ----------------------------------------------------------------------------
    case 0:
      mom  =  pow(abs(S0),2.) + pow(abs(P0),2.) + pow(abs(D0),2.);
      mom += 2.*real(-D1*conj(D1m) + D2*conj(D2m) - P1*conj(P1m));
      mom *= 2.;
      break;
      // L = 1 ----------------------------------------------------------------------------
    case 10:
      tmp  = 4./sqrt(3.)*P0*conj(S0+2./sqrt(5.)*D0);
      tmp -= 4./sqrt(5.)*(P1*conj(D1m) + P1m*conj(D1));
      mom  = real(tmp);
      break;
    case 11:
      tmp  = 2./sqrt(3.)*(P1-P1m)*conj(S0-D0/sqrt(5.));
      tmp += 2.*sqrt(2./5.)*(P1*conj(D2m) - P1m*conj(D2));
      tmp += 2.*sqrt(1./5.)*(P0*conj(D1) - P0*conj(D1m));
      mom  = real(tmp);
      break;
      // L = 2 ----------------------------------------------------------------------------
    case 20:
      mom =  4./sqrt(5.)*real(S0*conj(D0));
      mom += 4./7.*pow(abs(D0),2) + 4./5.*pow(abs(P0),2);
      mom -= 4./7.*real(D1*conj(D1m) + 2.*D2*conj(D2m));
      mom += 4./5.*real(P1*conj(P1m));
      break;
    case 21:
      tmp  = 2.*( D0/7. + S0/sqrt(5.) )*( conj(D1 - D1m) );
      tmp += 2./7.*sqrt(6.)*(D1*conj(D2m) - D1m*conj(D2));
      tmp += 2.*sqrt(3.)/5.*P0*conj(P1-P1m);
      mom  = real(tmp);
      break;
    case 22:
      tmp  = 2./35.*(7*sqrt(5.)*S0-10.*D0)*conj(D2+D2m);
      tmp += 1./7.*sqrt(6.)*(pow(abs(D1),2) + pow(abs(D1m),2));
      tmp += 1./5.*sqrt(6.)*(pow(abs(P1),2) + pow(abs(P1m),2));
      mom  = real(tmp);
      break;
      // L = 3 ----------------------------------------------------------------------------
    case 30:
      tmp  = 12./7./sqrt(5.)*( sqrt(3.)*D0*conj(P0) + P1*conj(D1m) + P1m*conj(D1) );
      mom  = real(tmp);
      break;
    case 31:
      tmp  = 4./7.*sqrt(6./5.)*P0*conj(D1-D1m) + 6./7.*sqrt(2./5.)*D0*conj(P1-P1m);
      tmp -= 2./7.*sqrt(3./5.)*(P1*conj(D2m) - P1m*conj(D2));
      mom  = real(tmp);
      break;
    case 32:
      tmp  = sqrt(12.)/7.*( P0*conj(D2+D2m) + sqrt(2.)*(P1*conj(D1) + P1m*conj(D1m)) );
      mom  = real(tmp);
      break;
    case 33:
      mom  = 6./7.*real(P1*conj(D2) - P1m*conj(D2m));
      break;
      // L = 4 ----------------------------------------------------------------------------
    case 40:
      mom  = 4./7.*pow(abs(D0),2);
      mom += 4./21.*real(4.*D1*conj(D1m) + D2*conj(D2m));
      break;
    case 41:
      mom  = 2.*sqrt(30.)/21.*real(D0*conj( D1 - D1m ));
      mom += 2.*sqrt(5.)/21.*real(D1m*conj(D2) - D1*conj(D2m));
      break;
    case 42:
      mom  = 2.*sqrt(15.)/21.*real(D0*conj( D2+D2m ));
      mom += 2*sqrt(10.)/21.*(pow(abs(D1),2.) + pow(abs(D1m),2.));
      break;
    case 43:
      mom = 2./3.*sqrt(5./7.)*real(D2*conj(D1) - D2m*conj(D1m));
      break;
    case 44:
      mom = 1./3.*sqrt(10./7.)*(pow(abs(D2),2.)+pow(abs(D2m),2.));
      break;
    default:
      mom = 0;
      break;
  }
  
  return mom;
}

double moments_H2_SPD(int L, int M, std::complex <double> *waves)
{
  /* return Im H2(LM) for one set of reflectivity waves
   * one needs to SUBTRACT the two reflectivities to get the full moments
   *      H2(LM) = ^(+)H2(LM) - ^(-)H2(LM)
   *
   * Waves ordering:
   * waves[0]         = S_0 ;
   * waves[1,2,3]     = P_{-1,0,+1} ;
   * waves[4,5,6,7,8] = D_{-2,-1,0,+1,+2}
   */
  
  double mom = 0;
  std::complex <double> tmp = 0;
  std::complex <double> S0 = waves[0];
  std::complex <double> P1m = waves[1], P0  = waves[2], P1 = waves[3];
  std::complex <double> D2m = waves[4], D1m = waves[5], D0 = waves[6],
  D1  = waves[7], D2  = waves[8];
  
  switch (10*L+M) {
      // L = 1 ----------------------------------------------------------------------------
    case 11:
      tmp  = -2./sqrt(3.)*(P1+P1m)*conj(S0-D0/sqrt(5.));
      tmp += 2.*sqrt(2./5.)*(P1*conj(D2m) + P1m*conj(D2));
      tmp -= 2.*sqrt(1./5.)*(P0*conj(D1m) + P0*conj(D1));
      mom  = real(tmp);
      break;
      // L = 2 ----------------------------------------------------------------------------
    case 21:
      tmp  = -2.*( D0/7. + S0/sqrt(5.) )*( conj(D1 + D1m) );
      tmp -= 2.*sqrt(3.)/5.*P0*conj(P1+P1m);
      tmp += 2./7.*sqrt(6.)*(D1*conj(D2m)+D1m*conj(D2));
      mom  = real(tmp);
      break;
    case 22:
      tmp  = -2./35.*(7*sqrt(5.)*S0-10.*D0)*conj(D2-D2m);
      tmp += sqrt(6.)/5.*(P1m*conj(P1m)-P1*conj(P1));
      tmp += 1./7.*sqrt(6.)*(D1m*conj(D1m)-D1*conj(D1));
      mom  = real(tmp);
      break;
      // L = 3 ----------------------------------------------------------------------------
    case 31:
      tmp  = -4./7.*sqrt(6./5.)*P0*conj(D1+D1m) - 6./7.*sqrt(2./5.)*D0*conj(P1+P1m);
      tmp -= 2./7.*sqrt(3./5.)*(P1*conj(D2m) + P1m*conj(D2));
      mom  = real(tmp);
      break;
    case 32:
      tmp  = -sqrt(12.)/7.*( P0*conj(D2-D2m) + sqrt(2.)*(P1*conj(D1) - P1m*conj(D1m)) );
      mom  = real(tmp);
      break;
    case 33:
      mom  = -6./7.*real(P1*conj(D2) + P1m*conj(D2m));
      break;
      // L = 4 ----------------------------------------------------------------------------
    case 41:
      mom  = -2.*sqrt(30.)/21.*real(D0*conj( D1 + D1m ));
      mom -= 2.*sqrt(5.)/21.*real(D1m*conj(D2) + D1*conj(D2m));
      break;
    case 42:
      mom  = -2.*sqrt(15.)/21.*real(D0*conj( D2-D2m ));
      mom -= 2*sqrt(10.)/21.*(pow(abs(D1),2.) - pow(abs(D1m),2.));
      break;
    case 43:
      mom = -2./3.*sqrt(5./7.)*real(D2*conj(D1) + D2m*conj(D1m));
      break;
    case 44:
      mom = -1./3.*sqrt(10./7.)*(pow(abs(D2),2.)-pow(abs(D2m),2.));
      break;
    default:
      mom = 0;
      break;
  }
  
  return mom;
}
